import pandas as pd
import os
from os.path import join as pjoin
import argparse
import logging
import sys
import pysam
from ugbio_core.logger import logger
import warnings

warnings.filterwarnings('ignore')


def run(argv):
    """
    converts CNV calls in bed format to vcf.
    input arguments:
    --cnv_annotated_bed_file: input bed file holding CNV calls.
    --fasta_index_file: (.fai file) tab delimeted file holding reference genome chr ids with their lengths.
    --out_directory: output directory
    --sample_name: sample name
    output files:
    vcf file: <sample_name>.cnv.vcf.gz
        shows called CNVs in zipped vcf format.
    vcf index file: <sample_name>.cnv.vcf.gz.tbi
        vcf corresponding index file.
    """
    parser = argparse.ArgumentParser(
        prog="cnv_results_to_vcf.py", description="converts CNV calls in bed format to vcf."
    )

    parser.add_argument("--cnv_annotated_bed_file", help="input bed file holding CNV calls", required=True, type=str)
    parser.add_argument("--fasta_index_file",
                        help="tab delimeted file holding reference genome chr ids with their lengths. (.fai file)",
                        required=True, type=str)
    parser.add_argument("--out_directory", help="output directory", required=False, type=str)
    parser.add_argument("--sample_name", help="sample name", required=True, type=str)
    parser.add_argument("--verbosity", help="Verbosity: ERROR, WARNING, INFO, DEBUG", required=False, default="INFO", )

    args = parser.parse_args(argv[1:])
    logger.setLevel(getattr(logging, args.verbosity))

    header = pysam.VariantHeader()

    # Add meta-information to the header
    header.add_meta('fileformat', value='VCFv4.2')
    header.add_meta('source', value='ULTIMA_CNV')

    # Add sample names to the header
    sample_name = args.sample_name
    header.add_sample(sample_name)

    header.add_line('##GENOOX_VCF_TYPE=ULTIMA_CNV')

    # Add contigs info to the header
    df_genome = pd.read_csv(args.fasta_index_file, sep='\t', header=None, usecols=[0, 1])
    df_genome.columns = ['chr', 'length']
    for index, row in df_genome.iterrows():
        chrID = row['chr']
        length = row['length']
        header.add_line(f"##contig=<ID={chrID},length={length}>")

    # Add ALT
    header.add_line('##ALT=<ID=<CNV>,Description="Copy number variant region">')
    header.add_line('##ALT=<ID=<DEL>,Description="Deletion relative to the reference">')
    header.add_line('##ALT=<ID=<DUP>,Description="Region of elevated copy number relative to the reference">')

    # Add FILTER
    header.add_line('##FILTER=<ID=PASS,Description="high confidence CNV call">')
    header.add_line('##FILTER=<ID=UG-CNV-LCR,Description="CNV calls overlpping (>50% overlap) with UG-CNV-LCR">')
    header.add_line('##FILTER=<ID=LEN,Description="CNV calls with length less then 10Kb">')

    # Add INFO
    header.add_line(
        '##INFO=<ID=CONFIDENCE,Number=1,Type=String,Description="Confidence level for CNV call.can be one of: LOW,MEDIUM,HIGH">')
    header.add_line('##INFO=<ID=CopyNumber,Number=1,Type=Float,Description="copy number of CNV call">')
    header.add_line('##INFO=<ID=RoundedCopyNumber,Number=1,Type=Integer,Description="rounded copy number of CNV call">')
    # header.add_line('##INFO=<ID=END_POS,Description="end position of the CNV">')
    header.add_line('##INFO=<ID=SVLEN,Number=.,Type=Integer,Description="CNV length">')
    header.add_line('##INFO=<ID=SVTYPE,Number=1,Type=String,Description="CNV type. can be DUP or DEL">')

    # Add FORMAT
    header.add_line('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">')

    # Open a VCF file for writing
    if args.out_directory:
        out_directory = args.out_directory
    else:
        out_directory = ""
    outfile = pjoin(out_directory, sample_name + '.cnv.vcf.gz')

    with pysam.VariantFile(outfile, mode='w', header=header) as vcf_out:
        df_cnvs = pd.read_csv(args.cnv_annotated_bed_file, sep="\t", header=None)
        df_cnvs.columns = ['chr', 'start', 'end', 'info']
        for index, row in df_cnvs.iterrows():
            # Create a new VCF record
            chrID = row['chr']
            start = row['start']
            end = row['end']
            info = row['info']

            CN = int(info.split("|")[0].replace("CN", ""))
            cnv_type = '<DUP>'
            if CN < 2:
                cnv_type = '<DEL>'

            filters = []
            for item in info.split(';'):
                arr = item.split('|')
                if len(arr) > 1:
                    filters.append(arr[1])

            record = vcf_out.new_record()
            record.contig = chrID
            record.start = start
            record.stop = end
            record.ref = "N"
            record.alts = (cnv_type,)

            CONFIDENCE = 'HIGH'
            if len(filters) > 0:
                for f in filters:
                    record.filter.add(f)
                CONFIDENCE = 'MEDIUM'
                if len(filters) > 1:
                    CONFIDENCE = 'LOW'
            else:
                record.filter.add('PASS')

            record.info['CONFIDENCE'] = CONFIDENCE
            record.info['CopyNumber'] = CN
            record.info['RoundedCopyNumber'] = int(round(CN))
            record.info['SVLEN'] = int(end) - int(start)
            record.info['SVTYPE'] = cnv_type.replace("<", "").replace(">", "")
            # record.info['END_POS'] = str(end)

            # Set genotype information for each sample
            GT = [None, 1]
            if CN == 1:
                GT = [0, 1]
            elif CN == 0:
                GT = [1, 1]
            record.samples[sample_name]['GT'] = (GT[0], GT[1])

            # Write the record to the VCF file
            vcf_out.write(record)

        vcf_out.close()

        cmd = f"tabix -p vcf {outfile}"
        os.system(cmd)
        logger.info(f"output file: {outfile}")
        logger.info(f"output file index: {outfile}.tbi")


if __name__ == "__main__":
    run(sys.argv)


# Copyright 2022 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Filter CNV calls by length and UG-CNV-LCR.
# CHANGELOG in reverse chronological order

import argparse
import logging
import os
import sys
import warnings

from ugbio_core.logger import logger
from ugbio_core import filter_bed

warnings.filterwarnings("ignore")

bedtools = "bedtools"
bedmap = "bedmap"


def annotate_bed(bed_file, lcr_cutoff, lcr_file, prefix, length_cutoff=10000):
    # get filters regions
    lcr_bed_file = filter_bed.filter_by_bed_file(bed_file, lcr_cutoff, lcr_file, prefix, "UG-CNV-LCR")
    length_bed_file = filter_bed.filter_by_length(bed_file, length_cutoff, prefix)

    # merge all filters and sort
    out_filters_unsorted = prefix + "filters.annotate.unsorted.bed"
    cmd = "cat " + lcr_bed_file + " " + length_bed_file + " > " + out_filters_unsorted

    os.system(cmd)
    out_filters_sorted = prefix + "filters.annotate.bed"

    cmd = bedtools + " sort -i " + out_filters_unsorted + " > " + out_filters_sorted

    os.system(cmd)
    out_bed_file_sorted = prefix + os.path.basename(bed_file).rstrip(".bed") + ".sorted.bed"
    cmd = bedtools + " sort -i " + bed_file + " > " + out_bed_file_sorted

    os.system(cmd)

    # annotate bed files by filters
    out_unsorted_annotate = prefix + os.path.basename(bed_file).rstrip(".bed") + ".unsorted.annotate.bed"
    cmd = (
        bedmap
        + " --echo --echo-map-id-uniq --delim '\\t' "
        + out_bed_file_sorted
        + " "
        + out_filters_sorted
        + " > "
        + out_unsorted_annotate
    )
    os.system(cmd)
    # combine to 4th column

    out_combined_info = prefix + os.path.basename(bed_file).rstrip(".bed") + ".unsorted.annotate.combined.bed"
    cmd = (
        "cat "
        + out_unsorted_annotate
        + ' | awk \'{if($5=="")'
        + "{print $0}"
        + 'else{print $1"\t"$2"\t"$3"\t"$5}}\' | sed \'s/\t$//\' > '
        + out_combined_info
    )
    os.system(cmd)

    out_annotate = prefix + os.path.basename(bed_file).rstrip(".bed") + ".annotate.bed"
    cmd = "sort -k1,1V -k2,2n -k3,3n " + out_combined_info + " > " + out_annotate
    os.system(cmd)

    out_filtered = prefix + os.path.basename(bed_file).rstrip(".bed") + ".filter.bed"
    cmd = "cat " + out_annotate + " | grep -v \"|\" | sed 's/\t$//' > " + out_filtered

    os.system(cmd)

    return [out_annotate, out_filtered]


def check_path(path):
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
        logger.info("creating out directory : %s", path)


def run(argv):
    """
    Given a bed file, this script will filter it by :
    1. lcr bed (ug_cnv_lcr) file
    3. length
    output consists of 2 files:
    - annotated bed file with filtering tags
    - filtered bed file
    """
    parser = argparse.ArgumentParser(
        prog="filter_sample_cnvs.py", description="Filter cnvs bed file by: ug_cnv_lcr, length"
    )
    parser.add_argument("--input_bed_file", help="input bed file with .bed suffix", required=True, type=str)
    parser.add_argument(
        "--intersection_cutoff",
        help="intersection cutoff for bedtools substruct function",
        required=True,
        type=float,
        default=0.5,
    )
    parser.add_argument("--cnv_lcr_file", help="UG-CNV-LCR bed file", required=True, type=str)
    parser.add_argument("--min_cnv_length", required=True, type=int, default=10000)
    parser.add_argument(
        "--out_directory",
        help="out directory where intermediate and output files will be saved."
        " if not supplied all files will be written to current directory",
        required=False,
        type=str,
    )
    parser.add_argument(
        "--verbosity",
        help="Verbosity: ERROR, WARNING, INFO, DEBUG",
        required=False,
        default="INFO",
    )

    args = parser.parse_args(argv[1:])
    logger.setLevel(getattr(logging, args.verbosity))

    prefix = ""
    if args.out_directory:
        prefix = args.out_directory
        prefix = prefix.rstrip("/") + "/"
    [out_annotate_bed_file, out_filtered_bed_file] = annotate_bed(
        args.input_bed_file, args.intersection_cutoff, args.cnv_lcr_file, prefix, args.min_cnv_length
    )

    logger.info("output files:")
    logger.info(out_annotate_bed_file)
    logger.info(out_filtered_bed_file)


if __name__ == "__main__":
    run(sys.argv)


from os.path import join as pjoin
import hashlib
import warnings

warnings.filterwarnings('ignore')

from . import get_resource_dir

from ugbio_cnv import convert_cnv_results_to_vcf


def compare_zipped_files(a, b):
    fileA = hashlib.sha256(open(a, 'rb').read()).digest()
    fileB = hashlib.sha256(open(b, 'rb').read()).digest()
    if fileA == fileB:
        return True
    else:
        return False


class TestConvertCnvResultsToVcf:
    inputs_dir = get_resource_dir(__file__)

    def test_convert_cnv_results_to_vcf(self, tmpdir):
        input_bed_file = pjoin(self.inputs_dir, "EL-0059.cnvs.annotate.bed")
        genome_file = pjoin(self.inputs_dir, "Homo_sapiens_assembly38.chr1-24.genome")
        expected_out_vcf = pjoin(self.inputs_dir, 'EL-0059.cnv.vcf.gz')

        sample_name = 'EL-0059'
        out_dir = f"{tmpdir}"
        convert_cnv_results_to_vcf.run([
            "convert_cnv_results_to_vcf",
            "--cnv_annotated_bed_file",
            input_bed_file,
            "--out_directory",
            out_dir,
            "--sample_name",
            sample_name,
            "--fasta_index_file",
            genome_file
        ])

        out_vcf_file = pjoin(tmpdir, sample_name + '.cnv.vcf.gz')
        assert compare_zipped_files(out_vcf_file, expected_out_vcf)
